"""
The `Metadata API <https://help.koordinates.com/api/publisher-admin-api/metadata-api/>`_
provides an interface for adding, inspecting and downloading
XML metadata documents against a range of objects.
"""

from requests_toolbelt.downloadutils import stream

from . import base
from . import exceptions


class MetadataManager(base.InnerManager):
    """
    Accessor for querying and updating metadata.

    Access via the ``metadata`` property of :py:class:`koordinates.layers.Layer`
    or :py:class:`koordinates.sets.Set` instances.
    """

    def set(self, parent_url, fp):
        """
        If the parent object already has XML metadata, it will be overwritten.

        Accepts XML metadata in any of the three supported formats.
        The format will be detected from the XML content.

        The Metadata object becomes invalid after setting

        :param file fp: A reference to an open file-like object which the content will be read from.
        """
        url = parent_url + self.client.get_url_path("METADATA", "POST", "set", {})
        r = self.client.request(
            "POST", url, data=fp, headers={"Content-Type": "text/xml"}
        )
        if r.status_code not in [200, 201]:
            raise exceptions.ServerError(
                "Expected success response, got %s: %s" % (r.status_code, url)
            )


class Metadata(base.InnerModel):
    FORMAT_ISO = "iso"
    FORMAT_FGDC = "fgdc"
    FORMAT_DC = "dc"
    FORMAT_NATIVE = "native"

    class Meta:
        manager = MetadataManager

    def get_xml(self, fp, format=FORMAT_NATIVE):
        """
        Returns the XML metadata for this source, converted to the requested format.
        Converted metadata may not contain all the same information as the native format.

        :param file fp: A path, or an open file-like object which the content should be written to.
        :param str format: desired format for the output. This should be one of the available
            formats from :py:meth:`.get_formats`, or :py:attr:`.FORMAT_NATIVE` for the native format.

        If you pass this function an open file-like object as the fp parameter, the function will
        not close that file for you.
        """
        r = self._client.request("GET", getattr(self, format), stream=True)
        filename = stream.stream_response_to_file(r, path=fp)
        return filename

    def get_formats(self):
        """ Return the available format names for this metadata """
        formats = []
        for key in (self.FORMAT_DC, self.FORMAT_FGDC, self.FORMAT_ISO):
            if hasattr(self, key):
                formats.append(key)
        return formats



# -*- coding: utf-8 -*-

"""
koordinates.client
==================
"""

import copy
import json
import logging
import os
import re
import sys

import pkg_resources
import requests
import requests_toolbelt
import six

from . import layers, licenses, publishing, sets, users, catalog, sources, exports
from . import exceptions


logger = logging.getLogger(__name__)


class Client(object):
    """
    A `Client` is used to define the host and api-version which the user
    wants to connect to. The user identity is also defined when `Client`
    is instantiated.
    """

    def __init__(self, host, token=None, activate_logging=False):
        """
        :param str host: the domain name of the Koordinates site to connect to (eg. ``labs.koordinates.com``)
        :param str token: Koordinates API token to use for authentication
        :param bool activate_logging: if True then logging to stderr is activated
        """
        if activate_logging:
            logging.basicConfig(
                stream=sys.stderr,
                level=logging.DEBUG,
                format="%(asctime)s %(levelname)s %(module)s %(message)s",
            )

        logger.debug("Initializing Client object for %s", host)

        self.host = host

        if token:
            self.token = token
        elif "KOORDINATES_TOKEN" in os.environ:
            self.token = os.environ["KOORDINATES_TOKEN"]
        else:
            raise KeyError(
                "No authentication token specified, and KOORDINATES_TOKEN not available in the environment."
            )

        self._user_agent = requests_toolbelt.user_agent(
            "KoordinatesPython", pkg_resources.require("koordinates")[0].version
        )

        self._init_managers(
            public={
                "sets": sets.SetManager,
                "publishing": publishing.PublishManager,
                "layers": layers.LayerManager,
                "tables": layers.TableManager,
                "licenses": licenses.LicenseManager,
                "catalog": catalog.CatalogManager,
                "sources": sources.SourceManager,
                "exports": exports.ExportManager,
            },
            private=(
                users.GroupManager,
                users.UserManager,
                sources.ScanManager,
                sources.DatasourceManager,
                exports.CropFeatureManager,
                exports.CropLayerManager,
            ),
        )

        self._session = requests.Session()
        self._session.headers.update(
            {"Accept": "application/json", "User-Agent": self._user_agent,}
        )
        if self.token:
            self._session.headers["Authorization"] = "key {token}".format(
                token=self.token
            )

        super(self.__class__, self).__init__()

    def _init_managers(self, public, private):
        self._manager_map = {}
        for manager_class in private:
            mgr = manager_class(self)
            self._register_manager(mgr.model, mgr)

        for alias, manager_class in public.items():
            mgr = manager_class(self)
            self._register_manager(mgr.model, mgr)
            setattr(self, alias, mgr)

    def _register_manager(self, model, manager):
        self._manager_map[model] = manager

    def get_manager(self, model):
        """
        Return the active manager for the given model.
        :param model: Model class to look up the manager instance for.
        :return: Manager instance for the model associated with this client.
        """
        if isinstance(model, six.string_types):
            # undocumented string lookup
            for k, m in self._manager_map.items():
                if k.__name__ == model:
                    return m
            else:
                raise KeyError(model)

        return self._manager_map[model]

    def _assemble_headers(self, method, user_headers=None):
        """
        Takes the supplied headers and adds in any which
        are defined at a client level and then returns
        the result.

        :param user_headers: a `dict` containing headers defined at the
                             request level, optional.

        :return: a `dict` instance
        """

        headers = copy.deepcopy(user_headers or {})

        if method not in ("GET", "HEAD"):
            headers.setdefault("Content-Type", "application/json")

        return headers

    def request(self, method, url, *args, **kwargs):
        headers = self._assemble_headers(method, kwargs.pop("headers", {}))
        r = self._raw_request(method, url, headers, *args, **kwargs)

        if method == "POST":
            # If we're posting to an endpoint
            if r.status_code in (requests.codes.created, requests.codes.accepted):
                # and we get a 201/202 response
                if "location" in r.headers:
                    # with a location header
                    logger.info("%s -> %s", r.status_code, r.headers["location"])

                    # follow it and return the results of the GET
                    r2 = self.request("GET", r.headers["location"])
                    r2.parent_response = r
                    return r2

        return r

    def _raw_request(self, method, url, headers, *args, **kwargs):
        # for the Koordinates library logging, strip auth tokens from log messages
        # and log POST/PUT bodies if we're sending JSON.
        # Get low-level logging via the requests.packages.urllib3 logger.
        if "json" in kwargs:
            logger.info(
                "Request: %s %s headers=%s body=%s",
                method,
                url,
                json.dumps(headers),
                json.dumps(kwargs["json"]),
            )
        else:
            logger.info("Request: %s %s headers=%s", method, url, json.dumps(headers))

        try:
            r = self._session.request(method, url, headers=headers, *args, **kwargs)
            logger.info("Response: %d %s in %s", r.status_code, r.reason, r.elapsed)
            logger.debug("Response: headers=%s", r.headers)
            r.raise_for_status()
            return r
        except requests.HTTPError as e:
            logger.warn("Response: %s: %s", e, r.text)
            raise exceptions.ServerError.from_requests_error(e)
        except requests.RequestException as e:
            raise exceptions.ServerError.from_requests_error(e)

    def get_url_path(self, datatype, verb, urltype, params={}, api_version=None):
        api_version = api_version or "v1"
        templates = getattr(self, "URL_TEMPLATES__%s" % api_version)

        url = templates[datatype][verb][urltype]
        return url.format(**params)

    def reverse_url(
        self, datatype, url, verb="GET", urltype="single", api_version=None
    ):
        """
        Extracts parameters from a populated URL

        :param datatype: a string identifying the data the url accesses.
        :param url: the fully-qualified URL to extract parameters from.
        :param verb: the HTTP verb needed for use with the url.
        :param urltype: an adjective used to the nature of the request.
        :return: dict
        """
        api_version = api_version or "v1"
        templates = getattr(self, "URL_TEMPLATES__%s" % api_version)

        # this is fairly simplistic, if necessary we could use the parse lib
        template_url = r"https://(?P<api_host>.+)/services/api/(?P<api_version>.+)"
        template_url += re.sub(
            r"{([^}]+)}", r"(?P<\1>.+)", templates[datatype][verb][urltype]
        )
        # /foo/{foo_id}/bar/{id}/
        m = re.match(template_url, url or "")
        if not m:
            raise KeyError(
                "No reverse match from '%s' to %s.%s.%s"
                % (url, datatype, verb, urltype)
            )

        r = m.groupdict()
        del r["api_host"]
        if r.pop("api_version") != api_version:
            raise ValueError("API version mismatch")
        return r

    def get_url(
        self, datatype, verb, urltype, params={}, api_host=None, api_version=None
    ):
        """Returns a fully formed url

        :param datatype: a string identifying the data the url will access.
        :param verb: the HTTP verb needed for use with the url.
        :param urltype: an adjective used to the nature of the request.
        :param \*\*params: substitution variables for the URL.
        :return: string
        :rtype: A fully formed url.
        """
        api_version = api_version or "v1"
        api_host = api_host or self.host

        subst = params.copy()
        subst["api_host"] = api_host
        subst["api_version"] = api_version

        url = "https://{api_host}/services/api/{api_version}"
        url += self.get_url_path(datatype, verb, urltype, params, api_version)
        return url.format(**subst)

    URL_TEMPLATES__v1 = {
        "LAYER": {
            "GET": {
                "single": "/layers/{id}/",
                "multi": "/layers/",
                "multidraft": "/layers/drafts/",
            },
            "POST": {
                "create": "/layers/",
                "update": "/layers/{layer_id}/versions/import/",
            },
            "DELETE": {"delete": "/layers/{id}/",},
        },
        "LAYER_VERSION": {
            "GET": {
                "single": "/layers/{layer_id}/versions/{version_id}/",
                "multi": "/layers/{layer_id}/versions/",
                "draft": "/layers/{layer_id}/versions/draft/",
                "published": "/layers/{layer_id}/versions/published/",
            },
            "POST": {
                "create": "/layers/{layer_id}/versions/",
                "import": "/layers/{layer_id}/versions/{version_id}/import/",
                "publish": "/layers/{layer_id}/versions/{version_id}/publish/",
            },
            "PUT": {"edit": "/layers/{layer_id}/versions/{version_id}/",},
            "DELETE": {"single": "/layers/{layer_id}/versions/{version_id}/",},
        },
        "SET": {
            "GET": {
                "single": "/sets/{id}/",
                "metadata": "/sets/{id}/metadata",
                "multi": "/sets/",
                "multidraft": "/sets/drafts/",
            },
            "POST": {"create": "/sets/",},
            "DELETE": {"delete": "/sets/{id}/",},
        },
        "SET_VERSION": {
            "GET": {
                "single": "/sets/{id}/versions/{version_id}/",
                "multi": "/sets/{id}/versions/",
                "draft": "/sets/{id}/versions/draft/",
                "published": "/sets/{id}/versions/published/",
            },
            "POST": {
                "create": "/sets/{id}/versions/",
                "publish": "/sets/{id}/versions/{version_id}/publish/",
            },
            "PUT": {"edit": "/sets/{id}/versions/{version_id}/",},
            "DELETE": {"single": "/sets/{id}/versions/{version_id}/",},
        },
        "CATALOG": {"GET": {"multi": "/data/", "latest": "/data/latest/",},},
        "PUBLISH": {
            "GET": {"single": "/publish/{id}/", "multi": "/publish/",},
            "POST": {"create": "/publish/",},
            "DELETE": {"single": "/publish/{id}/",},
        },
        "LICENSE": {
            "GET": {
                "single": "/licenses/{id}/",
                "multi": "/licenses/",
                "cc": "/licenses/{slug}/{jurisdiction}/",
            },
        },
        "METADATA": {
            # InnerManager, so relative to a parent object
            "GET": {"get": "metadata/",},
            "POST": {"set": "metadata/",},
        },
        "SOURCE": {
            "GET": {"single": "/sources/{id}/", "multi": "/sources/",},
            "POST": {"create": "/sources/",},
            "DELETE": {"single": "/sources/{id}/",},
        },
        "SCAN": {
            "GET": {
                "single": "/sources/{source_id}/scans/{scan_id}/",
                "multi": "/sources/{source_id}/scans/",
                "log": "/sources/{source_id}/scans/{scan_id}/log/",
                "all": "/sources/scans/",
            },
            "DELETE": {"cancel": "/sources/{source_id}/scans/{scan_id}/",},
            "POST": {"create": "/sources/{source_id}/scans/",},
        },
        "DATASOURCE": {
            "GET": {
                "multi": "/sources/{source_id}/datasources/",
                "single": "/sources/{source_id}/datasources/{datasource_id}/",
            }
        },
        # InnerManager, so relative to a parent object
        "PERMISSION": {
            "GET": {"multi": "permissions/", "single": "permissions/{permission_id}/",},
            "POST": {"single": "permissions/",},
            "PUT": {"multi": "permissions/",},
        },
        "EXPORT": {
            "GET": {"multi": "/exports/", "single": "/exports/{id}/",},
            "POST": {"create": "/exports/", "validate": "/exports/validate/",},
            "OPTIONS": {"options": "/exports/",},
            "DELETE": {"single": "/exports/{id}/",},
        },
        "CROPLAYER": {
            "GET": {
                "multi": "/exports/croplayers/",
                "single": "/exports/croplayers/{id}/",
            },
        },
        "CROPFEATURE": {
            "GET": {
                "multi": "/exports/croplayers/{croplayer_id}/cropfeatures/",
                "single": "/exports/croplayers/{croplayer_id}/cropfeatures/{cropfeature_id}/",
            },
        },
    }



# -*- coding: utf-8 -*-

"""
koordinates.layers
==================

For getting, editing and updating layers and tables, via
the `Layers & Tables API <https://help.koordinates.com/api/publisher-admin-api/layers-tables-api/>`_.

"""
import logging

from .utils import make_date
from . import base
from .licenses import License
from .metadata import Metadata, MetadataManager
from .permissions import PermissionObjectMixin
from .publishing import Publish
from .users import Group
from .utils import is_bound


logger = logging.getLogger(__name__)


class LayerManager(base.Manager):
    """
    Accessor for querying Layers & Tables.

    Access via the ``layers`` property of a :py:class:`koordinates.client.Client` instance.
    """

    _URL_KEY = "LAYER"

    def __init__(self, client):
        super(LayerManager, self).__init__(client)
        # Inner model managers
        self.versions = LayerVersionManager(client, self)
        self._data = LayerDataManager(client, self)
        self._metadata = MetadataManager(client, self)

    def list_drafts(self):
        """
        A filterable list views of layers, returning the draft version of each layer.
        If the most recent version of a layer or table has been published already,
        it won’t be returned here.
        """
        target_url = self.client.get_url("LAYER", "GET", "multidraft")
        return base.Query(self, target_url)

    def create(self, layer):
        """
        Creates a new layer.
        All attributes except ``name`` and ``data.datasources`` are optional.
        :return: the new draft version of the layer.
        """
        target_url = self.client.get_url("LAYER", "POST", "create")
        r = self.client.request("POST", target_url, json=layer._serialize())
        return layer._deserialize(r.json(), self)

    def list_versions(self, layer_id):
        """
        Filterable list of versions of a layer, always ordered newest to oldest.

        If the version’s source supports revisions, you can get a specific revision using
        ``.filter(data__source__revision=value)``. Specific values depend on the source type.
        Use ``data__source_revision__lt`` or ``data__source_revision__gte`` to filter
        using ``<`` or ``>=`` operators respectively.
        """
        target_url = self.client.get_url(
            "LAYER_VERSION", "GET", "multi", {"layer_id": layer_id}
        )
        return base.Query(
            self,
            target_url,
            valid_filter_attributes=("data",),
            valid_sort_attributes=(),
        )

    def get_version(self, layer_id, version_id, expand=[]):
        """
        Get a specific version of a layer.
        """
        target_url = self.client.get_url(
            "LAYER_VERSION",
            "GET",
            "single",
            {"layer_id": layer_id, "version_id": version_id},
        )
        return self._get(target_url, expand=expand)

    def get_draft(self, layer_id, expand=[]):
        """
        Get the current draft version of a layer.
        :raises NotFound: if there is no draft version.
        """
        target_url = self.client.get_url(
            "LAYER_VERSION", "GET", "draft", {"layer_id": layer_id}
        )
        return self._get(target_url, expand=expand)

    def get_published(self, layer_id, expand=[]):
        """
        Get the latest published version of this layer.
        :raises NotFound: if there is no published version.
        """
        target_url = self.client.get_url(
            "LAYER_VERSION", "GET", "published", {"layer_id": layer_id}
        )
        return self._get(target_url, expand=expand)

    def create_draft(self, layer_id):
        """
        Creates a new draft version.

        If anything in the data object has changed then an import will begin immediately.
        Otherwise to force a re-import from the previous sources call :py:meth:`koordinates.layers.LayerManager.start_import`.

        :rtype: Layer
        :return: the new version
        :raises Conflict: if there is already a draft version for this layer.
        """
        target_url = self.client.get_url(
            "LAYER_VERSION", "POST", "create", {"layer_id": layer_id}
        )
        r = self.client.request("POST", target_url, json={})
        return self.create_from_result(r.json())

    def start_import(self, layer_id, version_id):
        """
        Starts importing the specified draft version (cancelling any running import),
        even if the data object hasn’t changed from the previous version.
        """
        target_url = self.client.get_url(
            "LAYER_VERSION",
            "POST",
            "import",
            {"layer_id": layer_id, "version_id": version_id},
        )
        r = self.client.request("POST", target_url, json={})
        return self.create_from_result(r.json())

    def start_update(self, layer_id):
        """
        A shortcut to create a new version and start importing it.
        Effectively the same as :py:meth:`koordinates.layers.LayerManager.create_draft` followed by :py:meth:`koordinates.layers.LayerManager.start_import`.
        """
        target_url = self.client.get_url(
            "LAYER", "POST", "update", {"layer_id": layer_id}
        )
        r = self.client.request("POST", target_url, json={})
        return self.parent.create_from_result(r.json())

    def set_metadata(self, layer_id, version_id, fp):
        """
        Set the XML metadata on a layer draft version.

        :param file fp: file-like object to read the XML metadata from.
        :raises NotAllowed: if the version is already published.
        """
        base_url = self.client.get_url(
            "LAYER_VERSION",
            "GET",
            "single",
            {"layer_id": layer_id, "version_id": version_id},
        )
        self._metadata.set(base_url, fp)


class Layer(base.Model, PermissionObjectMixin):
    """
    Represents a version of a single Layer or Table.
    """

    class Meta:
        manager = LayerManager
        filter_attributes = (
            "kind",
            "public",
            "group",
            "license",
            "category",
            "geotag",
            "tag",
            "q",
            "created_at",
            "updated_at",
        )
        ordering_attributes = (
            "name",
            "created_at",
            "updated_at",
            "popularity",
        )
        serialize_skip = ("permissions",)
        deserialize_skip = ("permissions",)

    def _serialize(self, with_data=True):
        o = super(Layer, self)._serialize()
        if not with_data and ("data" in o):
            del o["data"]
        return o

    def _deserialize(self, data, manager):
        super(Layer, self)._deserialize(data, manager)
        self.group = (
            Group()._deserialize(data["group"], manager.client.get_manager(Group))
            if data.get("group")
            else None
        )
        self.data = (
            LayerData()._deserialize(data["data"], manager._data, self)
            if data.get("data")
            else None
        )
        self.version = (
            LayerVersion()._deserialize(data["version"], manager.versions, self)
            if data.get("version")
            else None
        )
        self.collected_at = (
            [make_date(d) for d in data["collected_at"]]
            if data.get("collected_at")
            else None
        )
        self.license = (
            License()._deserialize(data["license"], manager.client.get_manager(License))
            if data.get("license")
            else None
        )
        self.metadata = (
            Metadata()._deserialize(data["metadata"], manager._metadata, self)
            if data.get("metadata")
            else None
        )
        return self

    @property
    def is_published_version(self):
        """ Return if this version is the published version of a layer """
        pub_ver = getattr(self, "published_version", None)
        this_ver = getattr(self, "this_version", None)
        return this_ver and pub_ver and (this_ver == pub_ver)

    @property
    def is_draft_version(self):
        """ Return if this version is the draft version of a layer """
        pub_ver = getattr(self, "published_version", None)
        latest_ver = getattr(self, "latest_version", None)
        this_ver = getattr(self, "this_version", None)
        return (
            this_ver
            and latest_ver
            and (this_ver == latest_ver)
            and (latest_ver != pub_ver)
        )

    @is_bound
    def list_versions(self):
        """
        Filterable list of versions of a layer, always ordered newest to oldest.

        If the version’s source supports revisions, you can get a specific revision using
        ``.filter(data__source__revision=value)``. Specific values depend on the source type.
        Use ``data__source_revision__lt`` or ``data__source_revision__gte`` to filter
        using ``<`` or ``>=`` operators respectively.
        """
        target_url = self._client.get_url(
            "LAYER_VERSION", "GET", "multi", {"layer_id": self.id}
        )
        return base.Query(
            self._manager,
            target_url,
            valid_filter_attributes=("data",),
            valid_sort_attributes=(),
        )

    @is_bound
    def get_version(self, version_id, expand=[]):
        """
        Get a specific version of this layer
        """
        target_url = self._client.get_url(
            "LAYER_VERSION",
            "GET",
            "single",
            {"layer_id": self.id, "version_id": version_id},
        )
        return self._manager._get(target_url, expand=expand)

    @is_bound
    def get_draft_version(self, expand=[]):
        """
        Get the current draft version of this layer.
        :raises NotFound: if there is no draft version.
        """
        target_url = self._client.get_url(
            "LAYER_VERSION", "GET", "draft", {"layer_id": self.id}
        )
        return self._manager._get(target_url, expand=expand)

    @is_bound
    def get_published_version(self, expand=[]):
        """
        Get the latest published version of this layer.
        :raises NotFound: if there is no published version.
        """
        target_url = self._client.get_url(
            "LAYER_VERSION", "GET", "published", {"layer_id": self.id}
        )
        return self._manager._get(target_url, expand=expand)

    @is_bound
    def create_draft_version(self):
        """
        Creates a new draft version from this model content.

        If anything in the data object has changed then an import will begin immediately.
        Otherwise to force a re-import from the previous sources call :py:meth:`koordinates.layers.Layer.start_import`.

        :rtype: Layer
        :return: the new version
        :raises Conflict: if there is already a draft version for this layer.
        """
        target_url = self._client.get_url(
            "LAYER_VERSION", "POST", "create", {"layer_id": self.id}
        )
        r = self._client.request("POST", target_url, json={})
        return self._manager.create_from_result(r.json())

    @is_bound
    def start_import(self, version_id=None):
        """
        Starts importing this draft layerversion (cancelling any running import), even
        if the data object hasn’t changed from the previous version.

        :raises Conflict: if this version is already published.
        """
        if not version_id:
            version_id = self.version.id

        target_url = self._client.get_url(
            "LAYER_VERSION",
            "POST",
            "import",
            {"layer_id": self.id, "version_id": version_id},
        )
        r = self._client.request("POST", target_url, json={})
        return self._deserialize(r.json(), self._manager)

    @is_bound
    def start_update(self):
        """
        A shortcut to create a new version and start importing it.
        Effectively the same as :py:meth:`.create_draft_version` followed by :py:meth:`koordinates.layers.Layer.start_import`.

        :rtype: Layer
        :return: the new version
        :raises Conflict: if there is already a draft version for this layer.
        """
        target_url = self._client.get_url(
            "LAYER", "POST", "update", {"layer_id": self.id}
        )
        r = self._client.request("POST", target_url, json={})
        return self._manager.create_from_result(r.json())

    @is_bound
    def publish(self, version_id=None):
        """
        Creates a publish task just for this version, which publishes as soon as any import is complete.

        :return: the publish task
        :rtype: Publish
        :raises Conflict: If the version is already published, or already has a publish job.
        """
        if not version_id:
            version_id = self.version.id

        target_url = self._client.get_url(
            "LAYER_VERSION",
            "POST",
            "publish",
            {"layer_id": self.id, "version_id": version_id},
        )
        r = self._client.request("POST", target_url, json={})
        return self._client.get_manager(Publish).create_from_result(r.json())

    @is_bound
    def save(self, with_data=False):
        """
        Edits this draft layerversion.
        # If anything in the data object has changed, cancel any existing import and start a new one.

        :param bool with_data: if ``True``, send the data object, which will start a new import and cancel
            any existing one. If ``False``, the data object will *not* be sent, and no import will start.
        :raises NotAllowed: if the version is already published.
        """
        target_url = self._client.get_url(
            "LAYER_VERSION",
            "PUT",
            "edit",
            {"layer_id": self.id, "version_id": self.version.id},
        )
        r = self._client.request(
            "PUT", target_url, json=self._serialize(with_data=with_data)
        )
        return self._deserialize(r.json(), self._manager)

    @is_bound
    def delete_version(self, version_id=None):
        """
        Deletes this draft version (revert to published)

        :raises NotAllowed: if this version is already published.
        :raises Conflict: if this version is already deleted.
        """
        if not version_id:
            version_id = self.version.id

        target_url = self._client.get_url(
            "LAYER_VERSION",
            "DELETE",
            "single",
            {"layer_id": self.id, "version_id": version_id},
        )
        r = self._client.request("DELETE", target_url)
        logger.info("delete_version(): %s", r.status_code)

    @is_bound
    def delete_layer(self):
        """
        Delete this layer.
        """
        target_url = self._client.get_url("LAYER", "DELETE", "single", {"id": self.id})
        r = self._client.request("DELETE", target_url)
        logger.info("delete(): %s", r.status_code)

    @is_bound
    def set_metadata(self, fp, version_id=None):
        """
        Set the XML metadata on this draft version.

        :param file fp: file-like object to read the XML metadata from.
        :raises NotAllowed: if this version is already published.
        """
        if not version_id:
            version_id = self.version.id

        base_url = self._client.get_url(
            "LAYER_VERSION",
            "GET",
            "single",
            {"layer_id": self.id, "version_id": version_id},
        )
        self._manager._metadata.set(base_url, fp)

        # reload myself
        r = self._client.request("GET", base_url)
        return self._deserialize(r.json(), self._manager)


class LayerVersionManager(base.InnerManager):
    _URL_KEY = "LAYER_VERSION"


class LayerVersion(base.InnerModel):
    """
    Represents the ``version`` property of a :py:class:`.Layer` instance.
    """

    class Meta:
        manager = LayerVersionManager


class LayerDataManager(base.InnerManager):
    _URL_KEY = "DATA"


class LayerData(base.InnerModel):
    """
    Represents the ``data`` property of a :py:class:`.Layer` instance.
    """

    class Meta:
        manager = LayerDataManager


# Aliases
Table = Layer
TableManager = LayerManager



# -*- coding: utf-8 -*-

"""
koordinates.catalog
===================

The `Data Catalog API <https://help.koordinates.com/api/publisher-admin-api/data-catalog-api/>`_
is a read-only API for listing data from the Koordinates catalog.
For write access to the listed items, refer to the item specific models.

"""
import logging

from . import base
from .layers import Layer, Table
from .sets import Set

logger = logging.getLogger(__name__)


class CatalogManager(base.Manager):
    """
    Accessor for querying across the site via the Catalog API.

    Access via the ``catalog`` property of a :py:class:`koordinates.client.Client` instance.
    """

    _URL_KEY = "CATALOG"

    def get(self, *args, **kwargs):
        raise NotImplementedError(
            "No support for getting individual items via the Catalog API"
        )

    def list(self):
        """
        The published version of each layer, table, set, document or source.
        If something hasn’t been published yet, it won’t appear here.
        """
        return super(CatalogManager, self).list()

    def _get_item_class(self, url):
        """ Return the model class matching a URL """
        if "/layers/" in url:
            return Layer
        elif "/tables/" in url:
            return Table
        elif "/sets/" in url:
            return Set
        # elif '/documents/' in url:
        #     return Document
        else:
            raise NotImplementedError("No support for catalog results of type %s" % url)

    def create_from_result(self, result):
        try:
            klass = self._get_item_class(result["url"])
            obj = klass()
            return obj._deserialize(result, self.client.get_manager(klass))
        except NotImplementedError:
            # return as dict
            return result

    def list_latest(self):
        """
        A filterable list view of layers, tables, sets, documents and sources, similar to :py:meth:`koordinates.catalog.CatalogManager.list`.
        This returns the latest version of each item, regardless of whether or not it has been published.
        """
        target_url = self.client.get_url(self._URL_KEY, "GET", "latest")
        filter_attrs = self.model._meta.filter_attributes + ("version",)
        return base.Query(self, target_url, valid_filter_attributes=filter_attrs)


class CatalogEntry(base.Model):
    class Meta:
        manager = CatalogManager
        filter_attributes = (
            "kind",
            "public",
            "group",
            "license",
            "category",
            "geotag",
            "tag",
            "q",
            "created_at",
            "updated_at",
        )
        ordering_attributes = (
            "name",
            "created_at",
            "updated_at",
            "popularity",
        )

    def __init__(self, **kwargs):
        raise TypeError("CatalogEntry isn't meant to be instantiated")
